#ifndef EP_BITMAPFONT_H
#define EP_BITMAPFONT_H

#include "bitmapfont_glyph.h"
#include "generated/shinonome_gothic.h"
#include "generated/shinonome_mincho.h"
#include "generated/bitmapfont_rmg2000.h"
#include "generated/bitmapfont_ttyp0.h"
#include "generated/bitmapfont_wqy.h"

#endif // EP_BITMAPFONT_H
