/*
 * This file is part of EasyRPG Player.
 *
 * EasyRPG Player is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EasyRPG Player is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EasyRPG Player. If not, see <http://www.gnu.org/licenses/>.
 */

#include "instrumentation.h"
#include "utils.h"

#ifdef PLAYER_INSTRUMENTATION_VTUNE
__itt_domain* Instrumentation::domain = nullptr;
#endif

void Instrumentation::Init(const char* name) {
#ifdef PLAYER_INSTRUMENTATION_VTUNE
	assert(!domain);
#ifdef _WIN32
	domain = __itt_domain_create(Utils::ToWideString(name).c_str());
#else
	domain = __itt_domain_create(name);
#endif
#else
	(void)name;
#endif
}
