/*
 * This file is part of EasyRPG Player.
 *
 * EasyRPG Player is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EasyRPG Player is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EasyRPG Player. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This is displayed on the index-page of the doxygen documentation.
 */

/** \mainpage
 * The Player is the game interpreter of games made with RPG Maker 2000 or
 * 2003 editors. Technically, the player is the replacement of the
 * RPG_RT.exe program executable and runtime interpreter of RPG Maker,
 * distributed with games.<br />
 * This page provides a documentation of all the player classes and how they
 * interact with each other.
 */
